/* lib/browser/content/terminal - A library to process command terminal output in the browser. */ import browser from "../utilities/browser.js";
import modal from "../utilities/modal.js";
import network from "../utilities/network.js";
import util from "../utilities/util.js";
// cspell:words agenttype, arrowdown, arrowup, pagedown, pageup
/**
 * Interaction methods for the command terminal in the browser.
 * ```typescript
 * interface module_browserTerminal {
 *     content: () => [HTMLElement, HTMLElement];
 *     events: {
 *         close: (event:MouseEvent) => void;
 *         command: (event:KeyboardEvent) => void;
 *         keyInput: (event:KeyboardEvent) => void;
 *         keyOutput: (event:KeyboardEvent) => void;
 *         receive: (socketData:socketData) => void;
 *     };
 *     tools: {
 *         controlKeys: (event:KeyboardEvent, list:HTMLElement) => void;
 *         populate: (element:HTMLElement, logs:string[]) => void;
 *         send: (box:modal, command:string, autoComplete:boolean) => void;
 *     };
 * }
 * ``` */ const terminal = {
    content: function browser_content_terminal_content() {
        const footer = document.createElement("div"), cwd = document.createElement("p"), logs = document.createElement("ol"), label = document.createElement("label"), span = document.createElement("span"), textArea = document.createElement("textarea"), scroll = {
            position: -1,
            entries: []
        };
        logs.setAttribute("data-scroll", JSON.stringify(scroll));
        logs.setAttribute("class", "terminal-list");
        logs.onkeyup = terminal.events.keyOutput;
        cwd.setAttribute("class", "terminal-cwd");
        footer.setAttribute("class", "footer");
        label.setAttribute("class", "terminal");
        textArea.setAttribute("wrap", "hard");
        textArea.setAttribute("spellcheck", "false");
        textArea.onkeydown = terminal.events.command;
        textArea.onkeyup = terminal.events.keyInput;
        textArea.onmouseup = modal.events.footerResize;
        textArea.onblur = modal.events.textSave;
        span.appendText("Terminal command input");
        label.appendChild(span);
        label.appendChild(textArea);
        footer.appendChild(cwd);
        footer.appendChild(label);
        return [
            logs,
            footer
        ];
    },
    events: {
        close: function browser_content_terminal_close(event) {
            const box = event.target.getAncestor("box", "class");
            terminal.tools.send(box, "close-modal", false);
            modal.events.close(event);
        },
        command: function browser_content_terminal_command(event) {
            const key = event.key.toLowerCase(), target = event.target, box = target.getAncestor("box", "class"), id = box.getAttribute("id"), list = box.getElementsByClassName("terminal-list")[0], clearTarget = function browser_content_terminal_command() {
                browser.data.modals[id].text_value = "";
                browser.data.modals[id].historyIndex = browser.data.modals[id].history.length;
                target.value = "";
                network.configuration();
            };
            if (key === "c" && event.ctrlKey === true) {
                terminal.tools.send(box, "close-modal", false);
                terminal.tools.populate(list, [
                    ""
                ]);
                clearTarget();
                return;
            }
            if (key === "enter" && event.shiftKey === false) {
                const value = target.value, history = browser.data.modals[id].history;
                event.preventDefault();
                if (value === "clear") {
                    const scroll = {
                        position: -1,
                        entries: []
                    };
                    list.appendText("", true);
                    list.setAttribute("data-scroll", JSON.stringify(scroll));
                } else if (value === "") {
                    terminal.tools.populate(list, [
                        ""
                    ]);
                } else {
                    terminal.tools.send(box, value, false);
                }
                if (history[history.length - 1] !== value && value !== "") {
                    history.push(value);
                }
                clearTarget();
            }
        },
        keyInput: function browser_content_terminal_keyInput(event) {
            const key = event.key.toLowerCase(), target = event.target, value = target.value.replace(/^\s+/, "").replace(/\s+$/, ""), box = target.getAncestor("box", "class"), list = box.getElementsByClassName("terminal-list")[0], id = box.getAttribute("id"), history = browser.data.modals[id].history;
            let index = isNaN(browser.data.modals[id].historyIndex) ? browser.data.modals[id].history.length : browser.data.modals[id].historyIndex;
            event.preventDefault();
            if (key === "arrowup") {
                if (index > 0) {
                    index = index - 1;
                    target.value = history[index];
                    browser.data.modals[id].historyIndex = index;
                    network.configuration();
                }
                return;
            }
            if (key === "arrowdown") {
                if (index < history.length) {
                    index = index + 1;
                    if (index === history.length) {
                        target.value = browser.data.modals[id].text_value;
                    } else {
                        target.value = history[index];
                    }
                    browser.data.modals[id].historyIndex = index;
                    network.configuration();
                }
                return;
            }
            if (key === "insert") {
                terminal.tools.send(box, value, true);
                modal.events.textTimer(event);
                return;
            }
            if (key === "end" || key === "home" || key === "pagedown" || key === "pageup") {
                terminal.tools.controlKeys(event, list);
                return;
            }
            modal.events.textTimer(event);
        },
        keyOutput: function browser_content_terminal_keyOutput(event) {
            const list = event.target;
            terminal.tools.controlKeys(event, list);
        },
        receive: function browser_content_terminal_receive(socketData) {
            const data = socketData.data, write = function browser_content_terminal_receive_update(box) {
                if (box !== null) {
                    const cwd = box.getElementsByClassName("terminal-cwd")[0];
                    terminal.tools.populate(box.getElementsByClassName("terminal-list")[0], data.logs);
                    if (browser.data.modals[data.id] !== null && browser.data.modals[data.id].text_placeholder !== data.directory) {
                        browser.data.modals[data.id].text_placeholder = data.directory;
                        network.configuration();
                    }
                    cwd.appendText(data.directory, true);
                }
            };
            if (data.id === "all") {
                const terminals = document.getModalsByModalType("terminal"), each = function browser_content_terminal_each(element) {
                    if (element.dataset.agent === data.agentSource.agent && element.dataset.agenttype === data.agentSource.agentType) {
                        write(element);
                    }
                };
                terminals.forEach(each);
            } else if (data.autoComplete > -1) {
                const box = document.getElementById(data.id);
                if (box !== null) {
                    const textArea = box.getElementsByTagName("textarea")[0];
                    textArea.value = data.instruction;
                    textArea.selectionStart = data.autoComplete;
                    textArea.selectionEnd = data.autoComplete;
                }
            } else {
                const element = document.getElementById(data.id);
                write(element);
            }
        }
    },
    tools: {
        controlKeys: function browser_content_terminal_controlKeys(event, list) {
            const key = event.key.toLowerCase();
            if (key === "end") {
                const parent = list.parentNode;
                parent.scrollTo(parent.scrollLeft, parent.scrollHeight);
                return;
            }
            if (key === "home") {
                const parent1 = list.parentNode;
                parent1.scrollTo(parent1.scrollLeft, 0);
                return;
            }
            if (key === "pagedown" && event.shiftKey === true) {
                const items = list.getElementsByTagName("li"), parent2 = list.parentNode, scroll = JSON.parse(list.dataset.scroll), scrollPosition = scroll.position + 1, position = scroll.entries[scrollPosition] === undefined || items[scroll.entries[scrollPosition]] === undefined ? parent2.scrollHeight : items[scroll.entries[scrollPosition]].offsetTop;
                parent2.scrollTo(parent2.scrollLeft, position);
                scroll.position = scrollPosition < scroll.entries.length ? scroll.position = scrollPosition : scroll.entries.length;
                list.setAttribute("data-scroll", JSON.stringify(scroll));
                return;
            }
            if (key === "pagedown") {
                const parent3 = list.parentNode;
                parent3.scrollTo(parent3.scrollLeft, parent3.clientHeight + parent3.scrollTop);
                return;
            }
            if (key === "pageup" && event.shiftKey === true) {
                const items1 = list.getElementsByTagName("li"), parent4 = list.parentNode, scroll1 = JSON.parse(list.dataset.scroll), scrollPosition1 = scroll1.position - 1, position1 = scroll1.entries[scrollPosition1] === undefined || items1[scroll1.entries[scrollPosition1]] === undefined ? 0 : items1[scroll1.entries[scrollPosition1]].offsetTop;
                parent4.scrollTo(parent4.scrollLeft, position1);
                scroll1.position = scrollPosition1 < 0 ? -1 : scrollPosition1;
                list.setAttribute("data-scroll", JSON.stringify(scroll1));
                return;
            }
            if (key === "pageup") {
                const parent5 = list.parentNode, height = parent5.clientHeight, top = parent5.scrollTop, vertical = height > top ? 0 : top - height;
                parent5.scrollTo(parent5.scrollLeft, vertical);
                return;
            }
        },
        populate: function browser_content_terminal_populate(element, logs) {
            const items = element.getElementsByTagName("li").length, each = function browser_content_terminalPopulate_each(logItem) {
                let count = 0;
                const li = document.createElement("li"), names = {
                    "1": "bold",
                    "3": "italic",
                    "4": "underline",
                    "22": "no-bold",
                    "23": "no-italic",
                    "24": "no-underline",
                    "30": "black",
                    "31": "red",
                    "32": "green",
                    "33": "yellow",
                    "34": "blue",
                    "35": "purple",
                    "36": "cyan",
                    "37": "white",
                    "39": "no-color"
                }, ansi = /(\u001b\[\d{1,2}(;\d{1,2})*m)+/, list = function browser_content_terminalPopulate_list(result) {
                    let formatList = result.replace(/^\u001b\[/, "").replace(/m\u001b\[/g, "|").replace(/m$/, "").replace(/;/g, "|").split("|"), index = formatList.indexOf("0"), endString = "";
                    if (result === "\u001b[0m") {
                        return "\u001b[0m";
                    }
                    if (index > -1) {
                        if (index === 0) {
                            endString = end();
                            do {
                                formatList.splice(0, 1);
                                index = formatList.indexOf("0");
                            }while (index === 0)
                        }
                        if (index > 0) {
                            formatList = formatList.slice(0, index);
                        }
                    }
                    formatList.forEach(function browser_Content_terminalPopulate_list_each(value, index, list) {
                        list[index] = names[value];
                    });
                    count = count + 1;
                    return `${endString}<span class="${formatList.join(" ")}">`;
                }, end = function browser_content_terminalPopulate_end() {
                    if (count > 0) {
                        let output = "";
                        do {
                            count = count - 1;
                            output = output + "</span>";
                        }while (count > 0)
                        return output;
                    }
                    return "";
                };
                logItem = util.sanitizeHTML(logItem);
                do {
                    logItem = logItem.replace(ansi, list).replace(/\u001b\[0m/g, end);
                }while (ansi.test(logItem) === true)
                if (count > 0) {
                    logItem = logItem + "</span>";
                }
                // eslint-disable-next-line
                li.innerHTML = logItem;
                element.appendChild(li);
            }, oldScroll = JSON.parse(element.dataset.scroll), scroll = {
                position: oldScroll.entries.length,
                entries: oldScroll.entries
            }, parent = element.parentNode, scrollBottom = parent !== null && (parent.scrollTop === 0 || parent.scrollTop + parent.clientHeight === parent.scrollHeight);
            scroll.entries.push(logs.length + items);
            logs.forEach(each);
            element.setAttribute("data-scroll", JSON.stringify(scroll));
            if (scrollBottom === true) {
                parent.scrollTo(0, parent.scrollHeight);
            }
        },
        send: function browser_content_terminal_send(box, command, autoComplete) {
            // send close signal on modal close
            // capture c + ctrl - without alt or shift
            const agentType = box.dataset.agenttype, payload = {
                agentRequest: agentType === "device" ? {
                    agent: browser.data.hashDevice,
                    agentType: "device",
                    share: box.getAttribute("id")
                } : {
                    agent: browser.data.hashUser,
                    agentType: "user"
                },
                agentSource: {
                    agent: box.dataset.agent,
                    agentType: agentType
                },
                autoComplete: autoComplete === true ? box.getElementsByTagName("textarea")[0].selectionStart : -1,
                directory: box.getElementsByClassName("terminal-cwd")[0].innerHTML,
                id: box.getAttribute("id"),
                instruction: command,
                logs: [],
                target: "agentSource"
            };
            network.send(payload, "terminal");
        }
    }
};
export default terminal;
