/* lib/terminal/server/transmission/sender - Abstracts away the communication channel from the message. */ import deviceMask from "../services/deviceMask.js";
import settings from "../services/settings.js";
import transmit_ws from "./transmit_ws.js";
import vars from "../../utilities/vars.js";
/**
 * An abstraction to manage traffic output abstracted away from specific network protocols.
 * ```typescript
 * interface module_sender {
 *     agentQueue: (type:socketType, agent:string, payload:socketData) => void;  // If the agent is offline the message will be queued.
 *     broadcast : (payload:socketData, listType:agentType | "browser") => void; // Send a specified ata package to all agents of a given agent type.
 *     route     : (destination:agentCopy, socketData:socketData, callback:(socketData:socketData) => void) => void; // Automation to redirect data packages to a specific agent examination of a service identifier and agent data.
 *     send      : (data:socketData, agents:transmit_agents) => void;            // Send a specified data package to a specified agent
 * }
 * ``` */ const sender = {
    agentQueue: function terminal_server_transmission_sender_agentQueue(type, agent, payload) {
        const socket = transmit_ws.clientList[type][agent]; //console.log(type+" "+agent+" "+payload.service+" "+(socket===undefined?undefined:socket.status));
        if (socket !== undefined && socket !== null && (socket.status === "open" || socket.status === "pending")) {
            transmit_ws.queue(payload, socket, 1);
        } else if (vars.test.type === "" && (type === "device" || type === "user") && vars.settings[type][agent] !== undefined) {
            const service_exclusions = [
                "agent-hash",
                "agent-management",
                "agent-online",
                "agent-status",
                "copy-list-request",
                "copy-list",
                "copy-send-file",
                "cut",
                "error",
                "hash-share",
                "invite",
                "GET",
                "log",
                "response-no-action",
                "settings",
                "test-browser"
            ];
            if (service_exclusions.indexOf(payload.service) < 0) {
                if (vars.settings.queue[type][agent] === undefined) {
                    vars.settings.queue[type][agent] = [];
                }
                if (vars.settings.queue[type][agent].length > 0 && JSON.stringify(vars.settings.queue[type][agent][vars.settings.queue[type][agent].length - 1]) === JSON.stringify(payload)) {
                    return;
                }
                vars.settings.queue[type][agent].push(payload);
                const settingsData = {
                    settings: vars.settings.queue,
                    type: "queue"
                };
                settings({
                    data: settingsData,
                    service: "settings"
                });
            }
        }
    },
    // send to all agents of a given type
    broadcast: function terminal_server_transmission_sender_broadcast(payload, listType) {
        if (listType === "browser") {
            const list = Object.keys(transmit_ws.clientList.browser);
            list.forEach(function terminal_server_transmission_transmitWs_broadcast_each(agent) {
                transmit_ws.queue(payload, transmit_ws.clientList.browser[agent], 1);
            });
        } else {
            const list1 = Object.keys(vars.settings[listType]);
            let index = list1.length;
            if (index > 0) {
                do {
                    index = index - 1;
                    sender.agentQueue(listType, list1[index], payload);
                }while (index > 0)
            }
        }
    },
    // direct a data payload to a specific agent as determined by the service name and the agent details in the data payload
    route: function terminal_server_transmission_sender_route(destination, socketData, callback) {
        const payload = socketData.data, agent = payload[destination];
        if (agent.user === vars.settings.hashUser) {
            const deviceCallback = function terminal_server_transmission_sender_route_deviceCallback(device) {
                if (vars.settings.device[device] !== undefined) {
                    agent.device = device;
                }
                if (device === vars.settings.hashDevice) {
                    // same device
                    callback(socketData);
                } else {
                    sender.send(socketData, {
                        device: agent.device,
                        user: agent.user
                    });
                }
            };
            if (agent.device.length === 141) {
                deviceMask.unmask(agent.device, deviceCallback);
            } else {
                deviceCallback(deviceMask.resolve(agent));
            }
        } else {
            let count = 0;
            const maskCallback = function terminal_server_transmission_sender_route_maskCallback() {
                count = count + 1;
                if (count === 2) {
                    sender.send(socketData, {
                        device: agent.device,
                        user: agent.user
                    });
                }
            }, agentSelf = function terminal_server_transmission_sender_route_agentSelf(type) {
                if (payload[type] !== null && payload[type] !== undefined && payload[type].user === vars.settings.hashUser) {
                    if (payload[type].share === "") {
                        deviceMask.mask(payload[type], maskCallback);
                    } else {
                        payload[type].device = "";
                        maskCallback();
                    }
                } else {
                    maskCallback();
                }
            }, copyAgents = [
                "agentRequest",
                "agentSource",
                "agentWrite"
            ];
            copyAgents.splice(copyAgents.indexOf(destination), 1);
            agentSelf(copyAgents[0]);
            agentSelf(copyAgents[1]);
        }
    },
    // send a specified data package to a specified agent
    send: function terminal_server_transmission_sender_send(data, agents) {
        if (agents !== null) {
            if (agents.user === "browser") {
                sender.broadcast(data, "browser");
            } else {
                if (agents.user === vars.settings.hashUser) {
                    if (agents.device.length === 141) {
                        deviceMask.unmask(agents.device, function terminal_server_transmission_sender_send_unmask(actualDevice) {
                            sender.agentQueue("device", actualDevice, data);
                        });
                    } else {
                        sender.agentQueue("device", agents.device, data);
                    }
                } else {
                    sender.agentQueue("user", agents.user, data);
                }
            }
        }
    }
};
export default sender;
